import sys
import os
import subprocess as sbp
from mkdesigner.utils import time_stamp, prepare_cmd, call_log


class RefIndex(object):

    def __init__(self, args):
        self.args = args
        self.out = args.middir
        self.ref = '{}/ref/{}'.format(self.out, os.path.basename(args.ref))

    def run(self):
        print(time_stamp(),
              'Indexing reference fasta.',
              flush=True)

        cmd1 = 'samtools faidx {} \
                >> {}/log/samtools.log \
                2>&1'.format(self.ref, self.out)
        
        cmd2 = 'picard CreateSequenceDictionary R={} \
                >> {}/log/picard.log \
                2>&1'.format(self.ref, self.out)

        cmd1 = prepare_cmd(cmd1)
        cmd2 = prepare_cmd(cmd2)

        try:
            sbp.run(cmd1,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True,
                    check=True)
        except sbp.CalledProcessError:
            call_log(self.out, 'samtools', cmd1)
            sys.exit(1)

        try:
            sbp.run(cmd2,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True,
                    check=True)
        except sbp.CalledProcessError:
            call_log(self.out, 'picard', cmd2)
            sys.exit(1)

        print(time_stamp(),
              'Done.',
              flush=True)


import math
import pandas as pd
from mkdesigner.utils import read_vcf, time_stamp
import matplotlib.pyplot as plt

class VisualizeMarker(object):
    def __init__(self, vcf, out_png, fai):
        self.vcf = vcf
        self.png = out_png

        #Prepare chromosome information
        self.fai_data = []
        with open(fai, 'r') as f:
            for row in f:
                row = row.strip()
                self.fai_data.append(row.split('\t'))
        self.colnames = ['chr', 'len', 'A', 'B', 'C']
        self.fai_data = pd.DataFrame(self.fai_data, columns=self.colnames)
        self.fai_data['len'] = self.fai_data['len'].astype(int)

        #Read vcf file
        vcf_list = read_vcf(vcf)
        col = vcf_list[1]
        data = vcf_list[2]
        self.data = pd.DataFrame(data, columns=col)
        self.data['POS'] = self.data['POS'].astype(int)

    def run(self):
        print(time_stamp(),
              'Drawing positions of selected markers.',
              flush=True)
        
        #number of digits in the length of the longest chromosome
        digits = math.floor(math.log10(max(self.fai_data['len'])))
        standard = 10**(digits)
        #if the longest chr length is 23098790,
        #digits = 7
        #standard = 10000000

        if(max(self.fai_data['len']) / standard < 2):
            standard = standard / 5
        elif(max(self.fai_data['len']) / standard < 5):
            standard = int(standard / 2)
        #if the longest chr length is 23098790,
        #standard = 5000000 
        
        y_axis_at = range(0, standard*11, standard)
        y_axis_lab = []
        if(standard >= 100000):
            st_lab = standard/1000000
            sign = 'M'
        elif(standard >= 100):
            st_lab = standard/1000
            sign = 'K'
        else:
            st_lab = standard
            sign = 'bp'
        
        for i in range(11):
            y_axis_lab.append('{}{}'.format(round(st_lab * i, 1), sign))
        
        longest_len = max(self.fai_data['len'])


        # Create a figure
        fig = plt.figure(figsize=(5, 5), dpi=144)
        ax = fig.add_subplot(111,
                             xlim=[-1, len(self.fai_data['chr'])], 
                             xticks=range(len(self.fai_data['chr'])), 
                             xticklabels=self.fai_data['chr'],
                             xlabel="Chromosome",
                             ylim=[longest_len*1.05, -longest_len*0.05],
                             yticks=y_axis_at, 
                             yticklabels=y_axis_lab,
                             ylabel="Position")
        plt.subplots_adjust(left=0.15, right=0.95, bottom=0.15, top=0.95)
        plt.xticks(rotation=45)

        for i in range(len(self.fai_data['chr'])):
            #plot([x1, x2], [y1, y2])
            ax.plot([i, i], [0, self.fai_data['len'][i]], color="black") 
            data_select = self.data[self.data['#CHROM'] == self.fai_data['chr'][i]]
            for j in range(len(data_select)):
                pos = data_select['POS'].iloc[j]
                ax.plot([i-0.3, i+0.3], [pos, pos], color="black", linewidth=0.5)

        plt.xlim(-1, len(self.fai_data['chr']))
        plt.ylim(longest_len*1.05, -longest_len*0.05)

        # Save figure
        fig.savefig(self.png, dpi=144)

        print(time_stamp(),
              'Done.',
              flush=True)
        





from datetime import datetime


def time_stamp():
    return '[mkdesigner:{}]'.format(datetime.now().strftime('%Y-%m-%d %H:%M:%S'))

def prepare_cmd(cmd):
    #this function make the command arguments splitted by only space.
    return ' '.join(cmd.split())

def call_log(out_dir, name, cmd):
    print(time_stamp(), 
          '!!ERROR!! {}\n'.format(cmd), 
          flush=True)

    print('please check {}/log/{}.log\n'.format(out_dir, name))

#returns list consisted from 3 elements [header, colnames, data]
def read_vcf(vcf_file):
    header = []
    colnames = []
    data = []
    with open(vcf_file, 'r') as v:
        for row in v:
            row = row.strip()
            if row[0:2] == '##':
                header.append(row)
            elif row[0:1] == '#':
                colnames = row.split('\t')
            else:
                data.append(row.split('\t'))
    return [header, colnames, data]



import argparse
import sys
import os
from mkdesigner.__init__ import __version__

class Params(object):

    def __init__(self, program_name):
        self.program_name = program_name

    def set_options(self):
        if self.program_name == 'mkvcf':
            parser = self.mkvcf_options()
        elif self.program_name == 'mkprimer':
            parser = self.mkprimer_options()
        elif self.program_name == 'mkselect':
            parser = self.mkselect_options()

        if len(sys.argv) == 1:
            args = parser.parse_args(['-h'])
        else:
            args = parser.parse_args()
        return args
    
    def mkvcf_options(self):
        parser = argparse.ArgumentParser(description='MKDesigner version {}'.format(__version__),
                                         formatter_class=argparse.RawTextHelpFormatter)
        parser.usage = ('mkvcf -r <FASTA> -b <BAM_1> -b <BAM_2>... -n <name_1> -n <name_2>... -O <STRING>\n')

        # set options
        parser.add_argument('-r', '--ref',
                            action='store',
                            required=True,
                            type=str,
                            help='Reference fasta.',
                            metavar='')

        parser.add_argument('-b', '--bam',
                            action='append',
                            required=True,
                            type=str,
                            help=('Bam files for variant calling.\n'
                                  'e.g. -b bam1 -b bam2 ... \n'
                                  'You must use this option 2 times or more\n'
                                  'to get markers in following analysis.'),
                            metavar='')
        
        parser.add_argument('-n', '--name',
                            action='append',
                            required=True,
                            type=str,
                            help=('Variety name of each bam file.\n'
                                  'e.g. -n name_bam1 -n name_bam2 ... \n'
                                  'You must use this option same times\n'
                                  'as -b.'),
                            metavar='')
        
        parser.add_argument('-O', '--output',
                            action='store',
                            required=True,
                            type=str,
                            help=('Identical name (must be unique).\n'
                                  'This will be stem of output directory name.'),
                            metavar='')
                            
        parser.add_argument('--cpu',
                            action='store',
                            default=2,
                            type=int,
                            help=('Number of CPUs to use.\n'),
                            metavar='')
        
        # set version
        parser.add_argument('-v', '--version',
                            action='version',
                            version='%(prog)s {}'.format(__version__))
        return parser
        
    def mkprimer_options(self):
        parser = argparse.ArgumentParser(description='MKDesigner version {}'.format(__version__),
                                         formatter_class=argparse.RawTextHelpFormatter)
        parser.usage = ('mkprimer -r <FASTA> -V <VCF> -n1 <name1> -n2 <name2>\n'
                        '         -O <output name> --type <SNP or INDEL>\n'
                        '         [--target <Target position>]\n'
                        '         [--limit <INT>]\n'
                        '         [--mindep <INT>] [--maxdep <INT>]\n'
                        '         [--min_prodlen <INT>] [--max_prodlen <INT>]\n'
                        '         [--margin <INT>] [--max_distance <INT>]')

        # set options
        parser.add_argument('-r', '--ref',
                            action='store',
                            required=True,
                            type=str,
                            help='Reference fasta.',
                            metavar='')

        parser.add_argument('-V', '--vcf',
                            action='store',
                            required=True,
                            type=str,
                            help=('VCF file without filtering.\n'
                                  'Recommended to use VCF made by "mkvcf" command.\n'
                                  'VCF must contain GT and DP field.'),
                            metavar='')
        
        parser.add_argument('-n1', '--name1',
                            action='append',
                            required=True,
                            type=str,
                            help=('Variety name 1.\n'
                                  'Must match VCF column names.\n'
                                  'This parameter can be specified multiple times to design common markers for multiple varieties.\n'),
                            metavar='')
        
        parser.add_argument('-n2', '--name2',
                            action='append',
                            required=True,
                            type=str,
                            help=('Variety name 2.\n'
                                  'Must match VCF column names.\n'
                                  'This parameter can be specified multiple times to design common markers for multiple varieties.\n'),
                            metavar='')
        
        parser.add_argument('-O', '--output',
                            action='store',
                            required=True,
                            type=str,
                            help=('Identical name (must be unique).\n'
                                  'This will be stem of output directory name.'),
                            metavar='')
        
        parser.add_argument('-T', '--type',
                            action='store',
                            required=True,
                            choices=['SNP',  'INDEL'],
                            help=('Type of variants.\n'
                                  'SNP or INDEL are supported.'),
                            metavar='')
        
        parser.add_argument('-t', '--target',
                            action='append',
                            default=None,
                            type=str,
                            help=('Target position where primers designed/\n'
                                  'e.g. "chr01:1000000-3500000"\n'
                                  'If not specified, the program process whole genome.\n'
                                  'This parameter can be specified multiple times.'),
                            metavar='')
        
        parser.add_argument('-l', '--limit',
                            action='store',
                            default=10000,
                            type=int,
                            help=('The upper limit of the number of primer design attempts.\n'
                                  'A large number will take a long time to calculate,\n'
                                  'but a small number may miss useful markers.\n'
                                  'default: 10000'),
                            metavar='')

        parser.add_argument('--blast_timeout',
                            action='store',
                            default=60.0,
                            type=float,
                            help=('If the process of primer specificity checking by BLAST\n'
                                  'took time more than this parameter,\n'
                                  'the variants considered to be non-specific.\n'
                                  'default: 60.0 (sec)'),
                            metavar='')
        
        parser.add_argument('--mindep',
                            action='store',
                            default=2,
                            type=int,
                            help=('Variants with more depth than this\n'
                                  'are judged as valid mutations\ndefault: 2'),
                            metavar='')
        
        parser.add_argument('--maxdep',
                            action='store',
                            default=200,
                            type=int,
                            help=('Variants with less depth than this\n'
                                  'are judged as valid mutations\ndefault: 200'),
                            metavar='')
        
        parser.add_argument('--mismatch_allowed',
                            action='store',
                            default=5,
                            type=int,
                            help=('Primers with more mismatch than this\n'
                                  'are ignored in specificity check.\ndefault: 5'),
                            metavar='')  
        
        parser.add_argument('--mismatch_allowed_3_terminal',
                            action='store',
                            default=1,
                            type=int,
                            help=('Primers with more mismatch than this\n'
                                  'in 5 bases of 3\' terminal end\n'
                                  'are ignored in specificity check.\ndefault: 1'),
                            metavar='')  
        
        parser.add_argument('--unintended_prod_size_allowed',
                            action='store',
                            default=4000,
                            type=int,
                            help=('Primer pairs producing unintended PCR product which is shorter than this\n'
                                  'are ignored in specificity check.\ndefault: 4000'),
                            metavar='')  
        
        parser.add_argument('--min_prodlen',
                            action='store',
                            default=150,
                            type=int,
                            help=('Minimum PCR product length.default: 150'),
                            metavar='')
        
        parser.add_argument('--max_prodlen',
                            action='store',
                            default=280,
                            type=int,
                            help=('Maximam PCR product length.\ndefault: 280'),
                            metavar='')
        
        parser.add_argument('--opt_prodlen',
                            action='store',
                            default=180,
                            type=int,
                            help=('Optical PCR product length.\ndefault: 180'),
                            metavar='')

        parser.add_argument('--margin',
                            action='store',
                            default=5,
                            type=int,
                            help=('Minimum distance from 3\' terminal of primer to variant.\ndefault: 5'),
                            metavar='')
        
        parser.add_argument('--search_span',
                            action='store',
                            default=140,
                            type=int,
                            help=('Intervals to search for primers upstream and downstream variants.\ndefault: 140'),
                            metavar='')
        
        parser.add_argument('--primer_num_consider',
                            action='store',
                            default=3,
                            type=int,
                            help=('Primer number considering in primer3 software.\ndefault: 3'),
                            metavar='')

        parser.add_argument('--primer_min_size',
                            action='store',
                            default=18,
                            type=int,
                            help=('Minimum primer size\ndefault: 18'),
                            metavar='')

        parser.add_argument('--primer_max_size',
                            action='store',
                            default=26,
                            type=int,
                            help=('Maximum primer size\ndefault: 26'),
                            metavar='')

        parser.add_argument('--primer_opt_size',
                            action='store',
                            default=20,
                            type=int,
                            help=('Optical primer size\ndefault: 20'),
                            metavar='')
        
        parser.add_argument('--cpu',
                            action='store',
                            default=2,
                            type=int,
                            help=('Number of CPUs to use.\n'),
                            metavar='')
        # set version
        parser.add_argument('-v', '--version',
                            action='version',
                            version='%(prog)s {}'.format(__version__))
        return parser
        
    def mkselect_options(self):
        parser = argparse.ArgumentParser(description='MKDesigner version {}'.format(__version__),
                                         formatter_class=argparse.RawTextHelpFormatter)
        parser.usage = ('mkselect -i <FASTA Index file>\n'
                        '         -V <VCF with Primer> -n <INT>\n'
                        '         -O <STRING>\n'
                        '         [-t <Target position>]\n'
                        '         [-d <TSV with marker density infomation>]\n'
                        '         [--avoid_lowercase]\n')
        
        # set options
        parser.add_argument('-i', '--fai',
                            action='store',
                            required=True,
                            type=str,
                            help='Index file (.fai) of reference fasta.',
                            metavar='')

        parser.add_argument('-V', '--vcf',
                            action='store',
                            required=True,
                            type=str,
                            help=('VCF file with primers.\n'
                                  'This file must be made by "mkprimer" command.'),
                            metavar='')
        
        parser.add_argument('-n', '--num_marker',
                            action='store',
                            required=True,
                            type=int,
                            help=('Number of markers selected.'),
                            metavar='')
        
        parser.add_argument('-O', '--output',
                            action='store',
                            required=True,
                            type=str,
                            help=('Identical name (must be unique).\n'
                                  'This will be stem of output directory name.'),
                            metavar='')
        
        parser.add_argument('-t', '--target',
                            action='append',
                            default=None,
                            type=str,
                            help=('Target position where primers designed\n'
                                  'e.g. "chr01:1000000-3500000"\n'
                                  'This parameter can be specified multiple times.'),
                            metavar='')
        
        parser.add_argument('-d', '--density',
                            action='store',
                            default=None,
                            type=str,
                            help=('TSV file with marker density infomation..\n'
                                  'This file must be formatted as "test/density.tsv".'),
                            metavar='')
        
        parser.add_argument('--mindif',
                            action='store',
                            default=0,
                            type=int,
                            help=('Set minimum differences\n'
                                  'of PCR product length between alleles.\n'
                                  'For SNP marker, this must be 0.'),
                            metavar='')
        
        parser.add_argument('--maxdif',
                            action='store',
                            default=50,
                            type=int,
                            help=('For indel marker, set maximum differences\n'
                                  'of PCR product length between alleles.'),
                            metavar='')
        
        parser.add_argument('--avoid_lowercase',
                            action='store_true',
                            help=('Select only primers written by uppercase.\n'
                                  'Lowercase may mean repeat sequence.'))

        # set version
        parser.add_argument('-v', '--version',
                            action='version',
                            version='%(prog)s {}'.format(__version__))
        return parser
    
    def mkvcf_check_args(self, args):
        #Does a project file with the same name exist?
        if os.path.isdir('{}_mkvcf'.format(args.output)):
            sys.stderr.write(('  Output directory already exist.\n'
                              '  Please rename the --output.\n'))
            sys.exit(1)

        if not os.path.isfile('{}'.format(args.ref)):
            sys.stderr.write('  Input reference FASTA does not exist.\n')
            sys.exit(1)

        #Do BAM files exist?
        #Is the extentions of files designeated as BAM really '.bam' ?
        for input_name in args.bam:
                if not os.path.isfile('{}'.format(input_name)):
                    sys.stderr.write('  At least one of input BAM does not exist.\n')
                    sys.exit(1)
                ext = os.path.splitext(input_name)
                if ext[-1] != '.bam':
                    sys.stderr.write(('  Please check input BAM file "{}".\n'
                                      '  The extension of this file is not "bam".\n').format(input_name))
                    sys.exit(1)

        #Do the number of BAM and the number of names match?
        if len(args.bam) != len(args.name) :
            sys.stderr.write(('  Number of input BAM files is not'
                              '  matched the number of names.\n\n'))
            sys.exit(1)

        #Names must be unique.
        name_unique = set(args.name)
        if len(args.name) != len(name_unique) :
            sys.stderr.write(('  Variety names must not be duplicated.\n\n'))
            sys.exit(1)

    def mkprimer_check_args(self, args):
        #Does a project file with the same name exist?
        if os.path.isdir('{}_mkprimer'.format(args.output)):
            sys.stderr.write(('  Output directory already exist.\n'
                              '  Please rename the --output.\n'))
            sys.exit(1)
        if not os.path.isfile('{}'.format(args.vcf)):
            sys.stderr.write('  Input VCF does not exist.\n')
            sys.exit(1)
        if not os.path.isfile('{}'.format(args.ref)):
            sys.stderr.write('  Input reference FASTA does not exist.\n')
            sys.exit(1)

    def mkselect_check_args(self, args):
        #Does a directory with the same name exist?
        if os.path.isdir('{}_mkselect'.format(args.output)):
            sys.stderr.write(('  Output directory already exist.\n'
                              '  Please rename the --output.\n'))
            sys.exit(1)
        if not os.path.isfile('{}'.format(args.vcf)):
            sys.stderr.write('  Input VCF does not exist.\n')
            sys.exit(1)
        if not os.path.isfile('{}'.format(args.fai)):
            sys.stderr.write('  Input FASTA index does not exist.\n')
            sys.exit(1)
        if args.density != None:
            if not os.path.isfile('{}'.format(args.density)):
                sys.stderr.write(' Input file for adjusting marker density does not exist.\n')
                sys.exit(1)




import sys
import os
import subprocess as sbp
from mkdesigner.utils import time_stamp, prepare_cmd, call_log


class HaploCall(object):

    def __init__(self, args, num, is2nd=False):
        self.args = args
        self.out = args.middir
        self.name = args.name[num]
        self.is2nd = is2nd

        self.ref = '{}/ref/{}'.format(self.out, os.path.basename(args.ref))
        #self.bam = '{}/bam/{}'.format(self.out, os.path.basename(args.bam[num]))
        #230821 modified not to copy bam file.
        self.bam = args.bam[num]

        if is2nd:
            self.vcfdir = '{}/vcf_2nd/{}'.format(self.out, self.name)
        else: #1st
            self.vcfdir = '{}/vcf_1st/{}'.format(self.out, self.name)

    def run(self):
        if self.is2nd:
            print(time_stamp(),
                '2nd haplotype calling for "{}".'.format(self.name),
                flush=True)

            cmd1 = ':'

            cmd2 = 'gatk HaplotypeCaller \
                    -R {} -I {} -O {}_raw_variants.vcf \
                    --alleles {}/vcf_1st/Merged_filtered_variants_too_large_indel_removed.vcf \
                    >> {}/log/gatk.log 2>&1'.format(self.ref, self.bam, self.vcfdir, self.out, self.out)
        else: #1st
            print(time_stamp(),
                '1st haplotype calling for "{}".'.format(self.name),
                flush=True)

            cmd1 = 'samtools index {} \
                    >> {}/log/samtools.log \
                    2>&1'.format(self.bam, self.out)
            
            cmd2 = 'gatk HaplotypeCaller \
                    -R {} -I {} -O {}_raw_variants.vcf \
                    >> {}/log/gatk.log 2>&1'.format(self.ref, self.bam, self.vcfdir, self.out)

        cmd3 = 'gatk SelectVariants \
                -R {} -V {}_raw_variants.vcf \
                -select-type SNP \
                -O {}_raw_snps.vcf \
                >> {}/log/gatk.log 2>&1'.format(self.ref, self.vcfdir, self.vcfdir, self.out)

        cmd4 = 'gatk SelectVariants \
                -R {} -V {}_raw_variants.vcf \
                -select-type INDEL \
                -O {}_raw_indels.vcf \
                >> {}/log/gatk.log 2>&1'.format(self.ref, self.vcfdir, self.vcfdir, self.out)

        cmd5 = 'gatk VariantFiltration \
                -R {} -V {}_raw_snps.vcf \
                -O {}_filtered_snps.vcf \
                -filter-name "QD_filter" -filter "QD < 20.0" \
                -filter-name "FS_filter" -filter "FS > 60.0" \
                -filter-name "MQ_filter" -filter "MQ < 40.0" \
                -filter-name "SOR_filter" -filter "SOR > 4.0" \
                -filter-name "MQRankSum_filter" -filter "MQRankSum < -12.5" \
                -filter-name "ReadPosRankSum_filter" -filter "ReadPosRankSum < -8.0" \
                >> {}/log/gatk.log 2>&1'.format(self.ref, self.vcfdir, self.vcfdir, self.out)

        cmd6 = 'gatk VariantFiltration \
                -R {} -V {}_raw_indels.vcf \
                -O {}_filtered_indels.vcf \
                -filter-name "QD_filter" -filter "QD < 20.0" \
                -filter-name "FS_filter" -filter "FS > 200.0" \
                -filter-name "SOR_filter" -filter "SOR > 10.0" \
                >> {}/log/gatk.log 2>&1'.format(self.ref, self.vcfdir, self.vcfdir, self.out)

        cmd7 = 'bgzip -c {}_filtered_snps.vcf \
                > {}_filtered_snps.vcf.gz'.format(self.vcfdir, self.vcfdir)

        cmd8 = 'bcftools index {}_filtered_snps.vcf.gz \
                >> {}/log/bcftools.log 2>&1'.format(self.vcfdir, self.out)

        cmd9 = 'bgzip -c {}_filtered_indels.vcf \
                > {}_filtered_indels.vcf.gz'.format(self.vcfdir, self.vcfdir)

        cmd10 = 'bcftools index {}_filtered_indels.vcf.gz \
                 >> {}/log/bcftools.log 2>&1'.format(self.vcfdir, self.out)

        cmd11 = 'bcftools concat -o {}_filtered_variants.vcf.gz \
                 -a -O z \
                 {}_filtered_snps.vcf.gz \
                 {}_filtered_indels.vcf.gz \
                 >> {}/log/bcftools.log 2>&1'.format(self.vcfdir, self.vcfdir, self.vcfdir, self.out)

        cmd12 = 'bcftools index {}_filtered_variants.vcf.gz \
                 >> {}/log/bcftools.log 2>&1'.format(self.vcfdir, self.out)

        cmd1 = prepare_cmd(cmd1)
        cmd2 = prepare_cmd(cmd2)
        cmd3 = prepare_cmd(cmd3)
        cmd4 = prepare_cmd(cmd4)
        cmd5 = prepare_cmd(cmd5)
        cmd6 = prepare_cmd(cmd6)
        cmd7 = prepare_cmd(cmd7)
        cmd8 = prepare_cmd(cmd8)
        cmd9 = prepare_cmd(cmd9)
        cmd10 = prepare_cmd(cmd10)
        cmd11 = prepare_cmd(cmd11)
        cmd12 = prepare_cmd(cmd12)

        try:
            sbp.run(cmd1,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True,
                    check=True)
        except sbp.CalledProcessError:
            call_log(self.out, 'samtools', cmd1)
            sys.exit(1)

        try:
            sbp.run(cmd2,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True,
                    check=True)
        except sbp.CalledProcessError:
            call_log(self.out, 'gatk', cmd2)
            sys.exit(1)
        try:
            sbp.run(cmd3,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True,
                    check=True)
        except sbp.CalledProcessError:
            call_log(self.out, 'gatk', cmd3)
            sys.exit(1)

        try:
            sbp.run(cmd4,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True,
                    check=True)
        except sbp.CalledProcessError:
            call_log(self.out, 'gatk', cmd4)
            sys.exit(1)

        try:
            sbp.run(cmd5,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True,
                    check=True)
        except sbp.CalledProcessError:
            call_log(self.out, 'gatk', cmd5)
            sys.exit(1)

        try:
            sbp.run(cmd6,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True,
                    check=True)
        except sbp.CalledProcessError:
            call_log(self.out, 'gatk', cmd6)
            sys.exit(1)

        try:
            sbp.run(cmd7,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True,
                    check=True)
        except sbp.CalledProcessError:
            call_log(self.out, 'bgzip', cmd7)
            sys.exit(1)

        try:
            sbp.run(cmd8,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True,
                    check=True)
        except sbp.CalledProcessError:
            call_log(self.out, 'bcftools', cmd8)
            sys.exit(1)

        try:
            sbp.run(cmd9,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True,
                    check=True)
        except sbp.CalledProcessError:
            call_log(self.out, 'bgzip', cmd9)
            sys.exit(1)

        try:
            sbp.run(cmd10,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True,
                    check=True)
        except sbp.CalledProcessError:
            call_log(self.out, 'bcftools', cmd10)
            sys.exit(1)

        try:
            sbp.run(cmd11,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True,
                    check=True)
        except sbp.CalledProcessError:
            call_log(self.out, 'bcftools', cmd11)
            sys.exit(1)

        try:
            sbp.run(cmd12,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True,
                    check=True)
        except sbp.CalledProcessError:
            call_log(self.out, 'bcftools', cmd12)
            sys.exit(1)


        print(time_stamp(),
              'Done ({}).'.format(self.name),
              flush=True)



import csv
from multiprocessing import Pool
import os
import subprocess as sbp
import sys
import pandas as pd
import timeout_decorator # type: ignore
from mkdesigner.utils import call_log, prepare_cmd, read_vcf, time_stamp

class AddPrimerToVcf(object):

    additional_info = ['name','num','chr','L_seq','R_seq','L_pos','R_pos','L_TM','R_TM','product_size']
    header = []
    colnames = []
    data = []

    def __init__(self, args):
        self.args = args

        self.out = args.middir
        self.ref = '{}/ref/{}'.format(self.out, os.path.basename(args.ref))
        self.vcf = '{}/vcf/{}_selected.vcf'.format(self.out, os.path.splitext(os.path.basename(args.vcf))[0])
        self.cpu = args.cpu
        self.scope = args.scope
        self.margin = args.margin
        self.limit = args.limit
        self.timeout = args.timeout


        self.output_tmp = '{}/vcf/{}_selected_tmp.vcf'.format(self.out, os.path.splitext(os.path.basename(args.vcf))[0])
        self.output_vcf = '{}/vcf/{}_selected_primer_added.vcf'.format(self.out, os.path.splitext(os.path.basename(args.vcf))[0])

    def run(self):
        print(time_stamp(),
              'Making databases.',
              flush=True)
        #Make database files needed for blastn
        cmd0 = 'makeblastdb -in {} \
                -parse_seqids \
                -dbtype nucl \
                >> {}/log/blastn.log 2>&1'.format(self.ref, self.out)
        cmd0 = prepare_cmd(cmd0)
        try:
            sbp.run(cmd0,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True,
                    check=True)
        except sbp.CalledProcessError:
            call_log(self.out, 'blastn', cmd0)
            sys.exit(1) 
        print(time_stamp(),
              'Done.',
              flush=True)
        
        #Read input VCF
        vcf_list = read_vcf(self.vcf)
        self.header = vcf_list[0]
        self.colnames = vcf_list[1]
        self.data = vcf_list[2]
        self.header.append('##addprimertovcf.py_{}'.format(time_stamp()))
        self.data = pd.DataFrame(self.data, columns=self.colnames)
        nrow_all = len(self.data) #This is all candidate variants.
        print(time_stamp(),
              'Input VCF has {} variants.'.format(nrow_all),
              flush=True)
        
        print(time_stamp(),
              'Removing variants which are impossible to be markers.',
              flush=True)
        self.data = self.delete_impossible_variants(self.data)
        nrow = len(self.data)
        print(time_stamp(),
              '{} variants are survived.'.format(nrow),
              flush=True)

        print(time_stamp(),
              'Reducing variants to the specified number (--limit).',
              flush=True)
        self.data = self.reduce_variants(self.data)
        nrow = len(self.data)
        print(time_stamp(),
              '{} variants are selected for making primers.'.format(nrow),
              flush=True)

        print(time_stamp(),
              'Searching primers and checking their specificity '
              'using {} threads.'.format(self.cpu),
              flush=True)
        no_primer = 0
        non_specific = 0
        success = 0
        blast_to = 0
        #Multi processing
        ##############################################
        with Pool(self.cpu) as p:
            results = p.map(self.parallel, range(nrow))
            #Each process returns list,
            #[id(range(nrow)), exit status(0~2), Desciption added to INFO]
        #############################################
        for res in results:
            id = res[0]
            status = res[1]
            add_str = res[2]
            if status == 0:
                no_primer += 1
                self.data.at[self.data.index[id], 'FILTER'] = 'NO_PRIMER'
            elif status == 1:
                non_specific += 1
                self.data.at[self.data.index[id], 'FILTER'] = 'NON_SPECIFIC'
            elif status == 2:
                success += 1
                self.data.at[self.data.index[id], 'FILTER'] = 'PASS'
            elif status == 9:
                blast_to += 1
                self.data.at[self.data.index[id], 'FILTER'] = 'BLAST_TIMEOUT'
            self.data.at[self.data.index[id], 'INFO'] += add_str

        with open(self.output_vcf, 'w') as o:
            for h in self.header:
                o.write('{}\n'.format(h))
            writer = csv.writer(o, delimiter='\t')
            #writer.writerows(header)
            writer.writerow(self.colnames)
        
        self.data.to_csv(self.output_tmp, sep='\t', header=False, index=False)
        cmd1 = 'cat {} >> {}'.format(self.output_tmp, self.output_vcf)
        cmd1 = prepare_cmd(cmd1)
        try:
            sbp.run(cmd1,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True, check=True)
        except sbp.CalledProcessError:
            print(time_stamp(),
              'Error occured writing vcf data.',
              flush=True)
            sys.exit(1) 

        print(time_stamp(),
              'Selecting primer has been finished.\n\
               {} variants are selected from {} candidates.\n\
               (No primer: {} variants, Only non-specific primer: {} variants,\n\
               BLAST timeout: {} variants)'.format(success, nrow, no_primer, non_specific, blast_to),
              flush=True)
        print(time_stamp(),
              'Done.',
              flush=True)

    def delete_impossible_variants(self, dataframe):
        #Read INFO column of VCF data
        delete_row = []
        for i in range(len(dataframe)):
            info = dataframe['INFO'][i].split(';')
            for elem in info:
                elem_split = elem.split('=')
                if elem_split[0] == 'SPAN':
                    #'span' indicates the distance to the mutations
                    # that exist before and after the current variant.
                    span = elem_split[1].split(',')
                    span = list(map(int, span))
                    break

            #Delete if it is impossible to make primers because span is too small.
            min_span = self.args.primer_min_size + self.margin
            if span[0] <= min_span or span[1] <= min_span:
                delete_row.append(i)
        dataframe_pass = dataframe.drop(dataframe.index[delete_row])
        dataframe_pass = dataframe_pass.reset_index(drop=True)
        return dataframe_pass

    def reduce_variants(self, dataframe):
        #Make dataset for select markers
        intervals = [] 
        cur_chr = ''

        for i in range(len(dataframe)):
            if dataframe.at[dataframe.index[i], '#CHROM'] != cur_chr:
                #When chromosome of [i] is different to [i+1]
                intervals.append(10**20) #use 10**20 as instead of infinity
                cur_chr = dataframe.at[dataframe.index[i], '#CHROM']
            else:
                intervals.append(int(dataframe.at[dataframe.index[i], 'POS']) 
                                 - int(dataframe.at[dataframe.index[i-1], 'POS']))
        intervals.append(10**20) #length of list is len(data_s) + 1
        #data_s       0   1   2   3   4   5
        #intervals  0   1   2   3   4   5   6 #[0] and [6] are infinity

        #Decrease markers to a specified number (-n)
        use_index = list(range(len(dataframe)))

        while self.limit < len(use_index):
            shortest = intervals.index(min(intervals))

            if intervals[shortest - 1] < intervals[shortest + 1]:
                intervals[shortest-1]=intervals[shortest-1]+intervals[shortest]
                del use_index[shortest - 1]
            else:
                intervals[shortest+1]=intervals[shortest+1]+intervals[shortest]
                del use_index[shortest]

            del intervals[shortest]
        
        dataframe_pass = dataframe.iloc[use_index]
        dataframe_pass = dataframe_pass.reset_index(drop=True)
        return dataframe_pass

    def parallel(self, i):
        #[0]:row number [1]:status(0 means NO_PRIMER, 1 means NON_SPECIFIC, 2 means PASS, 9 means BLAST_TIMEOUT) [2]description for adding primer infomation
        return_list = [i, 0, ';PRIMER=']

        series = self.data.iloc[i, :]
        st_in = self.samtools_input(series)
        st_out = self.samtools(st_in, i)
        primer3_format_file = self.make_primer3_input(series, st_out)
        '''
        if primer3_format_file == 'N':
            return_list = [i, 0, ';PRIMER=']
            return return_list
        '''
        primer3_result_file = self.primer3(primer3_format_file)
        blastn_data_list = self.make_blastn_input(primer3_result_file)
        if len(blastn_data_list) <= 1:
            return_list = [i, 0, ';PRIMER=']
            return return_list
        blastn_query_file = blastn_data_list[0]
        df_primer_info = blastn_data_list[1]

        #241023 set timeout
        try:
            blastn_result_file = self.blastn(blastn_query_file, self.timeout)
        except timeout_decorator.TimeoutError:
            return_list = [i, 9, ';PRIMER=']
            return return_list
        
        df_filtered = self.filter_blastn_result(blastn_result_file, df_primer_info['L_seq'], df_primer_info['R_seq'])
        if df_filtered.shape[0] <= 1:
            return_list = [i, 1, ';PRIMER=']
            return return_list
        use_primer_set = self.select_primer_set(df_primer_info, df_filtered)
        if use_primer_set == -1:
            return_list = [i, 1, ';PRIMER=']
            return return_list
        else:
            add_list = df_primer_info.iloc[use_primer_set].values.tolist()
            add_list = map(str, add_list)
            add = ';PRIMER={}'.format('|'.join(add_list))
            return_list = [i, 2, add]
            return return_list

    def samtools_input(self, series):
        chr = str(series['#CHROM'])
        ref_len = len(str(series['REF']))
        start = int(series['POS']) - self.scope
        end = int(series['POS']) + (ref_len - 1) + self.scope
        return_str = '{}:{}-{}'.format(chr, start, end)
        return return_str

    def samtools(self, st_in, i):
        cmd = 'samtools faidx {} {}'.format(self.ref, st_in, self.out, i)
        cmd = prepare_cmd(cmd)
        try:
            str = sbp.run(cmd, capture_output=True, text=True,
                          shell=True, check=True).stdout
        except sbp.CalledProcessError:
            call_log(self.out, 'samtools', cmd)
            sys.exit(1)
        str_tmp = str.split('\n')
        str_list = [str_tmp[0][1:], ''.join(str_tmp[1:])]
        return str_list
    
    def make_primer3_input(self, series, st_out):
        #Read INFO column of VCF data
        info = series['INFO'].split(';')
        for elem in info:
            elem_split = elem.split('=')
            if elem_split[0] == 'SPAN':
                span = elem_split[1].split(',')
                span = list(map(int, span))
                break

        #Modified to remove impossible variants before multi processing. 230809
        '''
        #if it is impossible to make primers because span is too small, primer3 input file is not made.
        min_span = self.args.primer_min_size + self.args.margin
        if span[0] <= min_span or span[1] <= min_span:
            return 'N'
        '''

        #Make parameters for primer3
        target_start = self.scope - self.margin + 1
        var_len = len(series['REF'])
        target_length = var_len + self.margin * 2
        #Avoid designing primers at surrounding variants.
        exclude = ''
        if span[0] <= self.scope:
            exclude += '{},{}'.format(1, self.scope - span[0] + 1)
        if span[1] <= self.scope + var_len - 1:
            exclude += ' {},{}'.format(self.scope + span[1], self.scope - span[1] + var_len)
        #For SNP markers, designated product length is used.
        #For INDEL markers, product length is determined by 
        # an expression with the size of INDEL as a variable.
        if self.args.type == 'SNP' :
            max_prd = self.args.args.max_prodlen
            opt_prd = self.args.args.opt_prodlen
            min_prd = self.args.args.min_prodlen
        elif self.args.type == 'INDEL' :
            X = abs(len(series['REF']) - len(series['ALT']))
            if len(series['REF']) > len(series['ALT']) :
                ref_shorter = 0
                alt_shorter = X
            else :
                ref_shorter = X
                alt_shorter = 0

            if X < 5 :
                max_prd = 100 - ref_shorter
                opt_prd = 75 - ref_shorter
            else:
                max_prd = round((300 * X) / (10 + X) - ref_shorter)
                opt_prd = round((max_prd * 0.75) - ref_shorter)
            min_prd = 50 + alt_shorter

        primer3_data_list = [
            'SEQUENCE_ID={}\n'.format(st_out[0]),
            'SEQUENCE_TEMPLATE={}\n'.format(st_out[1]),
            'SEQUENCE_TARGET={},{}\n'.format(target_start, target_length),
            'SEQUENCE_EXCLUDED_REGION={}\n'.format(exclude),
            'PRIMER_NUM_RETURN={}\n'.format(self.args.args.primer_num_consider),
            'PRIMER_OPT_SIZE={}\n'.format(self.args.args.primer_opt_size),
            'PRIMER_MIN_SIZE={}\n'.format(self.args.args.primer_min_size),
            'PRIMER_MAX_SIZE={}\n'.format(self.args.args.primer_max_size),
            'PRIMER_PRODUCT_OPT_SIZE={}\n'.format(opt_prd),
            'PRIMER_PRODUCT_SIZE_RANGE={}-{}\n'.format(min_prd, max_prd),
            'PRIMER_OPT_GC_PERCENT=50.0\n',
            'PRIMER_MIN_GC=30.0\n',
            'PRIMER_MAX_GC=70.0\n',
            'PRIMER_MIN_TM=55.0\n',
            'PRIMER_OPT_TM=61.0\n',
            'PRIMER_MAX_TM=67.0\n',
            'PRIMER_PAIR_MAX_DIFF_TM=5.0\n',
            'PRIMER_MAX_SELF_ANY=5\n',
            'PRIMER_MAX_SELF_END=2\n',
            'PRIMER_MAX_POLY_X=4\n',
            'PRIMER_EXPLAIN_FLAG=1\n',
            '=\n']
        primer3_data = ''.join(primer3_data_list)

        name_format = '{}/intermediate/format_{}_{}.txt'.format(self.out, series['#CHROM'], series['POS'])
        with open(name_format, 'w') as o:
            o.write(primer3_data)

        return name_format

    def primer3(self, input):
        name_result = str(input).replace('/format_', '/result_')
        cmd = 'primer3_core --output {} {} \
                >> {}/log/primer3.log 2>&1'.format(name_result, input, self.out)
        cmd = prepare_cmd(cmd)
        try:
            sbp.run(cmd,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True,
                    check=True)
        except sbp.CalledProcessError:
            call_log(self.out, 'primer3', cmd)
            sys.exit(1)

        os.remove(input)
        return name_result

    def make_blastn_input(self, input):
        sequence_id = ''
        chr = ''
        start_pos = 0
        idnum = 0
        primer_info = pd.DataFrame(columns=self.additional_info)
        with open(input, 'r') as inp:
            for line in inp:
                tmp = line.split('=')
                if len(tmp) < 2:
                    continue
                if tmp[0] == 'SEQUENCE_ID':
                    sequence_id = tmp[1].strip()
                    chr = sequence_id.split(':')[0]
                    start_pos = int(sequence_id.split(':')[1].split('-')[0])
                if tmp[0] == 'PRIMER_LEFT_{}_SEQUENCE'.format(idnum):
                    primer_row = pd.Series(index=self.additional_info, name=str(idnum), dtype=str)
                    primer_row['name'] = sequence_id
                    primer_row['num'] = idnum
                    primer_row['chr'] = chr
                    primer_row['L_seq'] = tmp[1].strip()
                if tmp[0] == 'PRIMER_RIGHT_{}_SEQUENCE'.format(idnum):
                    primer_row['R_seq'] = tmp[1].strip()
                if tmp[0] == 'PRIMER_LEFT_{}'.format(idnum):
                    L_pos = int(tmp[1].split(',')[0])
                    primer_row['L_pos'] = start_pos + L_pos
                if tmp[0] == 'PRIMER_RIGHT_{}'.format(idnum):
                    R_pos = int(tmp[1].split(',')[0])
                    primer_row['R_pos'] = start_pos + R_pos
                if tmp[0] == 'PRIMER_LEFT_{}_TM'.format(idnum):
                    primer_row['L_TM'] = tmp[1].strip()
                if tmp[0] == 'PRIMER_RIGHT_{}_TM'.format(idnum):
                    primer_row['R_TM'] = tmp[1].strip()
                if tmp[0] == 'PRIMER_PAIR_{}_PRODUCT_SIZE'.format(idnum):
                    primer_row['product_size'] = tmp[1].strip()
                    primer_info = pd.concat([primer_info, pd.DataFrame([primer_row])])
                    idnum += 1
        
        os.remove(input)
        if idnum == 0:
            return ''
        
        name_query = str(input).replace('/result_', '/query_')
        with open(name_query, 'w') as o:
            for j in range(idnum):
                o.write('>{}:{}-{}_{}\n'.format(chr, primer_info.at[primer_info.index[j], 'L_pos'], primer_info.at[primer_info.index[j], 'R_pos'], j))
                o.write('{}NNNNNNNNNNNNNNNNNNNN{}\n'.format(primer_info.at[primer_info.index[j], 'L_seq'], primer_info.at[primer_info.index[j], 'R_seq']))
        
        return_list = [name_query, primer_info]
        return return_list

    def blastn(self, input, time):
        @timeout_decorator.timeout(time) # type: ignore
        def blastn_inner(self, input):
            name_result = str(input).replace('/query_', '/blastn_')
            cmd = 'blastn -db {} -query {} -out {} \
                    -outfmt "6 std qseq sseq sstrand" \
                    -evalue 30000 -word_size 7 \
                    -num_alignments 50000 \
                    -penalty -1 -reward 1 \
                    -ungapped >> {}/log/blastn.log 2>&1'.format(self.ref, input,name_result, self.out)
            cmd = prepare_cmd(cmd)
            try:
                sbp.run(cmd,
                        stdout=sbp.DEVNULL,
                        stderr=sbp.DEVNULL,
                        shell=True, check=True)
            except sbp.CalledProcessError:
                call_log(self.out, 'blastn', cmd)
                sys.exit(1)

            os.remove(input)
            return name_result
        return blastn_inner(self, input)

    def filter_blastn_result(self, input, primers_L, primers_R):
        try:
            df = pd.read_csv(input, header=None, sep='\t')
        except pd.errors.EmptyDataError:
            os.remove(input)
            return pd.DataFrame()
        os.remove(input)
        col = ['qaccver', 'saccver', 'pident', 'length', 'mismatch', 'gapopen', 'qstart', 'qend', 'sstart', 'send', 'evalue', 'bitscore', 'qseq', 'sseq', 'sstrand']
        #Example: [test:98-322_0	test	100	20	0	0	41	60	322	303	0.002	33.3	TGGTCCTTCGAGTGGATGGA	TGGTCCTTCGAGTGGATGGA	minus]
        df.columns = col
        nrow = len(df)
        df['idnum'] = range(nrow) #Add id number to last column

        idnum = -1
        selected_row_index = []
        for j in range(nrow):
            while int(df['qaccver'][j].split('_')[-1]) != idnum:
                idnum += 1
                p_len_L = len(primers_L.iloc[idnum])
                p_len_R = len(primers_R.iloc[idnum])
                p_pos_L = [1, p_len_L]
                p_pos_R = [p_len_L + 20 + 1, p_len_L + 20 + p_len_R]

            df.at[df.index[j], 'idnum'] = idnum
            
            ## 'mm' means mismuch nucleotide number
            ## 'mm_3t' means mismuch nucleotide number in 3' terminal
            if df['qend'][j] < p_pos_R[1]/2: 
                #5' terminal
                if df['qstart'][j] - p_pos_L[0] < 2:
                    mm = df['qstart'][j] - p_pos_L[0]
                else:
                    mm = df['qstart'][j] - p_pos_L[0] - 1
                #If 2 or more bases from the end are not aligned, all are not considered mismatches and 1 is subtracted from the mismatch.
                #3' terminal
                if p_pos_L[1] - df['qend'][j] < 2:
                    mm3t = p_pos_L[1] - df['qend'][j]
                else:
                    mm3t = p_pos_L[1] - df['qend'][j] - 1
                mm = mm + mm3t
            else:
                #5' terminal
                if df['qstart'][j] - p_pos_R[0] < 2:
                    mm = df['qstart'][j] - p_pos_R[0]
                else:
                    mm = df['qstart'][j] - p_pos_R[0] - 1
                #3' terminal
                if p_pos_L[1] - df['qend'][j] < 2:
                    mm3t = p_pos_R[1] - df['qend'][j]
                else:
                    mm3t = p_pos_R[1] - df['qend'][j] - 1
                mm = mm + mm3t

            mm = mm + df['mismatch'][j] # add internal mismuch number
            if mm > self.args.args.mismatch_allowed:
                continue
            if mm3t > self.args.args.mismatch_allowed_3_terminal:
                continue
            
            qseq = df['qseq'][j][-5:]
            sseq = df['sseq'][j][-5:]
            for k in range(5-mm3t):
                if qseq[4-k] != sseq[4-k]:
                    mm3t += 1
            if mm3t > self.args.args.mismatch_allowed_3_terminal:
                continue
            
            #241023 This code may takes a lot of time...
            '''
            if df_selected is None:
                df_selected = pd.DataFrame([df.iloc[j]])
            else:
                df_selected = pd.concat([df_selected, pd.DataFrame([df.iloc[j]])])
            '''
            selected_row_index.append(j)

        if len(selected_row_index) == 0:
            df_selected = pd.DataFrame(columns=df.columns)
        else:
            df_selected = df.iloc[selected_row_index]
            df_selected = df_selected.reset_index(drop=True)
        
        df_selected.sort_values(by=['idnum','saccver','sstart'], inplace=True) #sort
        
        #Detecting unintended PCR products
        output_row_index = []
        #df_output = None
        for j in range(len(df_selected)):
            if df_selected.at[df_selected.index[j], 'sstrand'] == 'plus':
                pri = df_selected.at[df_selected.index[j], 'qaccver']
                chr = df_selected.at[df_selected.index[j], 'saccver']
                pos = df_selected.at[df_selected.index[j], 'sstart']
                cnt = 0
                while True:
                    cnt += 1
                    if j + cnt >= len(df_selected): break
                    if df_selected.at[df_selected.index[j+cnt], 'qaccver'] != pri: break
                    if df_selected.at[df_selected.index[j+cnt], 'saccver'] != chr: break
                    if df_selected.at[df_selected.index[j+cnt], 'sstrand'] == 'minus':
                        if df_selected.at[df_selected.index[j+cnt], 'sstart'] - pos < self.args.args.unintended_prod_size_allowed:
                            output_row_index.append(j)
                            output_row_index.append(j+cnt)
                            '''
                            if df_output is None:
                                df_output = pd.DataFrame([df_selected.iloc[j]])
                            else:
                                df_output = pd.concat([df_output, pd.DataFrame([df_selected.iloc[j]])])
                            df_output = pd.concat([df_output, pd.DataFrame([df_selected.iloc[j+cnt]])])
                            '''
                        else:
                            break
        if len(output_row_index) == 0:
            df_output = pd.DataFrame(columns=df_selected.columns)
        else:
            df_output = df_selected.iloc[output_row_index]
            df_output = df_output.reset_index(drop=True)
             
        name_blastn_filtered = str(input).replace('/blastn_', '/blastn_filtered_')
        df_output.to_csv(name_blastn_filtered, sep='\t', header=False, index=False)
        df_output.reset_index(drop=True)

        return df_output

    def select_primer_set(self, df_primer_info, df_blastn_info):
        use_primer = -1 #"-1" means "No specific primer set"
        #If Useful primer set was found, this parameter will be natural number.
        for j in range(len(df_primer_info)):
            id = '{}:{}-{}_{}'.format(
                    df_primer_info.at[df_primer_info.index[j], 'chr'],
                    df_primer_info.at[df_primer_info.index[j], 'L_pos'],df_primer_info.at[df_primer_info.index[j], 'R_pos'],df_primer_info.at[df_primer_info.index[j], 'num'])
            hit_cnt = 0
            match_cnt = 0
            for k in range(len(df_blastn_info)):
                if id == str(df_blastn_info.at[df_blastn_info.index[k], 'qaccver']):
                    hit_cnt += 1
                    if df_blastn_info.at[df_blastn_info.index[k], 'sstrand'] == 'plus':
                        if df_primer_info.at[df_primer_info.index[j], 'L_pos'] == df_blastn_info.at[df_blastn_info.index[k], 'sstart']:
                            match_cnt += 1
                    else:
                        if df_primer_info.at[df_primer_info.index[j], 'R_pos'] == df_blastn_info.at[df_blastn_info.index[k], 'sstart']:
                            match_cnt += 1
            if hit_cnt == 2 and match_cnt == 2:
                use_primer = j
                break
        return use_primer






import sys
import subprocess as sbp
from mkdesigner.utils import time_stamp, prepare_cmd, call_log

#This script is no longer used

class MvInputFiles(object):

    def __init__(self, args):
        self.args = args
        self.out = args.project

    def run(self):
        print(time_stamp(),
              'Moving input files to the working directory.',
              flush=True)

        cmd1 = 'cp {} {}/ref/'.format(self.args.ref, self.out)
        
        bams = ' '.join(self.args.bam)
        cmd2 = 'cp {} {}/bam/'.format(bams, self.out)

        cmd1 = prepare_cmd(cmd1)
        cmd2 = prepare_cmd(cmd2)

        try:
            sbp.run(cmd1,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True,
                    check=True)
        except sbp.CalledProcessError:
            print('Error occored at moving input files to the working directory.',flush=True)
            sys.exit(1)

        try:
            sbp.run(cmd2,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True,
                    check=True)
        except sbp.CalledProcessError:
            print('Error occored at moving input files to the working directory.',flush=True)
            sys.exit(1)

        print(time_stamp(),
              'Done.',
              flush=True)


#!/usr/bin/env python3

import sys
from mkdesigner.params import Params

pm = Params('mkvcf')
args = pm.set_options()


import os
from multiprocessing import Pool
from mkdesigner.utils import prepare_cmd, time_stamp
from mkdesigner.refindex import RefIndex
from mkdesigner.haplocall import HaploCall
from mkdesigner.mergevcf import MergeVcf
#230826 inserted
from mkdesigner.removetoolargeindel import RemoveTooLargeIndel
import subprocess as sbp

class MKVcf(object):

    def __init__(self, args):
        pm.mkvcf_check_args(args)
        self.args = args
        self.ref = args.ref
        self.bam = args.bam
        self.N_bam = len(args.bam)
        self.name = args.name
        self.cpu = args.cpu

        #240626 modified
        #devided output directory to {output}_mkvcf/ and {output}_mkvcf/intermediate/ 
        self.stem = args.output
        self.dir = '{}_mkvcf'.format(self.stem)
        self.middir = '{}_mkvcf/intermediate'.format(self.stem)

    def mkdir(self):
        os.mkdir('{}'.format(self.dir))
        os.mkdir('{}'.format(self.middir))
        os.mkdir('{}/log'.format(self.middir))
        os.mkdir('{}/ref'.format(self.middir))
        #230821 modified not to copy bam files.
        #os.mkdir('{}/bam'.format(self.proj))
        os.mkdir('{}/vcf_1st'.format(self.middir))
        os.mkdir('{}/vcf_2nd'.format(self.middir))

    def command(self):
        #Output command info
        command = ' '.join(sys.argv)
        fn = '{}/command.txt'.format(self.dir)
        with open(fn, 'w') as f:
            f.write('{}\n'.format(command))

    def mvinputfiles(self):
        cmd1 = 'cp {} {}/ref/'.format(self.args.ref, self.middir)
        cmd1 = prepare_cmd(cmd1)
        try:
            sbp.run(cmd1, stdout=sbp.DEVNULL, stderr=sbp.DEVNULL,
                    shell=True, check=True)
        except sbp.CalledProcessError:
            print('Error occored at moving input files to the working directory.',flush=True)
            sys.exit(1)

    def refindex(self):
        ri = RefIndex(self)
        ri.run()

    def haplocall_1st(self, num):
        hc = HaploCall(self, num, False)
        hc.run()

    def mergevcf_1st(self):
        mv = MergeVcf(self, False)
        mv.run()

    def remove_too_large_indel(self):
        rt = RemoveTooLargeIndel(self)
        rt.run()
        #Removing indels larger than 100 bp to avoid errors.

    def haplocall_2nd(self, num):
        hc = HaploCall(self, num, True)
        hc.run()

    def mergevcf_2nd(self):
        mv = MergeVcf(self, True)
        mv.run()

    def mvoutputfiles(self):
        cmd1 = 'cp {}/vcf_2nd/Merged_filtered_variants.vcf {}/{}_ready_for_mkprimer.vcf'.format(self.middir, self.dir, self.stem)
        cmd1 = prepare_cmd(cmd1)
        try:
            sbp.run(cmd1, stdout=sbp.DEVNULL, stderr=sbp.DEVNULL,
                    shell=True, check=True)
        except sbp.CalledProcessError:
            print('Error occored at moving output files.',flush=True)
            sys.exit(1)

def main():
    print(time_stamp(), 'MKVcf started.', flush=True)

    prog = MKVcf(args)
    prog.mkdir()
    prog.command()
    prog.mvinputfiles()
    prog.refindex()
    with Pool(prog.cpu) as p:
        p.map(prog.haplocall_1st, range(prog.N_bam))
    prog.mergevcf_1st()
    prog.remove_too_large_indel()
    with Pool(prog.cpu) as p:
        p.map(prog.haplocall_2nd, range(prog.N_bam))
    prog.mergevcf_2nd()
    prog.mvoutputfiles()
    
    print(time_stamp(), 'MKVcf successfully finished.\n', flush=True)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import csv
import os
import pandas as pd # type: ignore
import sys
import subprocess as sbp
from mkdesigner.utils import read_vcf, time_stamp, prepare_cmd
from mkdesigner.params import Params
from mkdesigner.visualize_marker import VisualizeMarker

pm = Params('mkselect')
args = pm.set_options()

class MKSelect(object):
    def __init__(self, args):
        pm.mkselect_check_args(args)
        self.args = args
        self.vcf = args.vcf
        self.fai = args.fai
        self.num_marker = args.num_marker
        self.target = args.target
        self.density = args.density
        self.mindif = args.mindif
        self.maxdif = args.maxdif

        self.stem = args.output
        self.dir = '{}_mkselect'.format(self.stem)
        os.mkdir(self.dir)

        #VCF data
        self.header = [] #VCF header
        self.colnames = [] #VCF colnames
        self.data = [] #VCF content
        self.data_s = [] #VCF content selected
        self.primers = [] #Primer information of selected variants
        self.out_table = [] #Output table, easy to read
        if self.density is None:
            self.den_table = None
        else:
            self.den_table = pd.read_csv(self.density, sep='\t') 
        #Marker density data

    def command(self):
        #Output command info
        command = ' '.join(sys.argv)
        fn = '{}/command.txt'.format(self.dir)
        with open(fn, 'w') as f:
            f.write('{}\n'.format(command))

    def readvcf(self):
        #Read input VCF
        vcf_list = read_vcf(self.vcf)
        self.header = vcf_list[0]
        self.colnames = vcf_list[1]
        self.data = vcf_list[2]
        self.header.append('##mkselect.py_{},num_primer={},target={}'.format(time_stamp(), self.num_marker, self.target))
        self.data = pd.DataFrame(self.data, columns=self.colnames)

    def filtervcf(self):
        #Select variants FILTER = PASS
        data_pass = self.data[self.data['FILTER'] == 'PASS']

        #Avoid lowercase
        if self.args.avoid_lowercase:
            print(time_stamp(), 'Removing markers with primers containing lowercase sequences.', flush=True)
            #Correct infomation of primers
            check_primers = self.get_info_primers(data_pass)
            #col = ['name','num','chr','L_seq','R_seq','L_pos','R_pos','L_TM','R_TM','product_size']
            delete_row = [] #row indexes to delete
            for i in range(len(check_primers)):
                if check_primers['L_seq'][i].isupper() and \
                   check_primers['R_seq'][i].isupper():
                    #if all characters are uppercase, TRUE
                    pass
                else:
                    delete_row.append(i)
            data_pass = data_pass.drop(data_pass.index[delete_row])
            
        #Select variants by --mindif and --maxdif
        str_len_ref = data_pass['REF'].str.len()
        str_len_alt = data_pass['ALT'].str.len()
        dif = str_len_ref - str_len_alt
        dif_abs = dif.abs()
        data_pass = data_pass[(dif_abs >= self.mindif) & (dif_abs <= self.maxdif)]
        
        #Reset index number (row)
        data_pass = data_pass.reset_index(drop=True)
        
        #Select variants if target position is designated.
        if self.target is not None:
            pass_row = [] #row indexes passed

            for h in range(len(self.target)):
                tmp = self.target[h].split(':')
                if len(tmp) != 2:
                    print(time_stamp(), 'input of "--target" is invalid.', flush=True)
                    sys.exit(1)
                tmp2 = tmp[1].split('-')
                if len(tmp2) != 2:
                    print(time_stamp(), 'input of "--target" is invalid.', flush=True)
                    sys.exit(1)
                target_chr = tmp[0]
                target_start = int(tmp2[0])
                target_end = int(tmp2[1])
                
                for i in range(len(data_pass)):
                    if data_pass['#CHROM'][i] == target_chr and \
                    int(data_pass['POS'][i]) > target_start and \
                    int(data_pass['POS'][i]) < target_end:
                        pass_row.append(i)

            #remove duplication and sort
            pass_row = list(set(pass_row))
            pass_row.sort()

            #Select only designated position
            data_pass = data_pass.iloc[pass_row]
            #Reset index number (row)
            data_pass = data_pass.reset_index(drop=True)

        #Make dataset for select markers
        intervals = [] 
        cur_chr = ''

        for i in range(len(data_pass)):
            if data_pass.at[data_pass.index[i], '#CHROM'] != cur_chr:
                #When chromosome of [i] is different to [i+1]
                intervals.append(10**20) #use 10**20 as instead of infinity
                cur_chr = data_pass.at[data_pass.index[i], '#CHROM']
                if self.den_table is not None:
                    cur_den = self.den_table[self.den_table['chr'] == cur_chr]
                    cur_den = cur_den.reset_index(drop=True)
            else:
                real_pos = [int(data_pass.at[data_pass.index[i-1], 'POS'])
                            , int(data_pass.at[data_pass.index[i], 'POS'])]

                #When there is no density information
                if self.den_table is None:
                    intervals.append(real_pos[1] - real_pos[0])
                else: #with density information, adjust position
                    adj_pos = []
                    for pos in real_pos:
                        new_pos = 0
                        for j in range(len(cur_den)):
                            st = int(cur_den.at[cur_den.index[j], 'start']) - 1
                            en = int(cur_den.at[cur_den.index[j], 'end'])
                            dn = float(cur_den.at[cur_den.index[j], 'density'])
                            if st < pos:
                                if pos < en:
                                    new_pos = new_pos + int((pos - st) * dn)
                                    break
                                else:
                                    new_pos = new_pos + int((en - st) * dn)
                        adj_pos.append(new_pos)
                    intervals.append(adj_pos[1] - adj_pos[0])

        intervals.append(10**20) #length of list is len(data_s) + 1
        #data_s       0   1   2   3   4   5
        #intervals  0   1   2   3   4   5   6 #[0] and [6] are infinity

        #Decrease markers to a specified number (-n)
        use_index = list(range(len(data_pass)))

        while self.num_marker < len(use_index):
            shortest = intervals.index(min(intervals))

            if intervals[shortest - 1] < intervals[shortest + 1]:
                intervals[shortest-1]=intervals[shortest-1]+intervals[shortest]
                del use_index[shortest - 1]
            else:
                intervals[shortest+1]=intervals[shortest+1]+intervals[shortest]
                del use_index[shortest]

            del intervals[shortest]
        
        self.data_s = data_pass.iloc[use_index]
        self.data_s = self.data_s.reset_index(drop=True)

        #Correct infomation of primers
        self.primers = self.get_info_primers(self.data_s)

    def get_info_primers(self, vcfdata):
        primers_list = []
        for i in range(len(vcfdata)):
            info = str(vcfdata.at[vcfdata.index[i], 'INFO'])
            info_spl = info.split(';')
            for j in range(len(info_spl)):
                elem = info_spl[j].split('=')
                if elem[0] == 'PRIMER':
                    pri = elem[1].split('|')
                    primers_list.append(pri)
        primers_col = ['name','num','chr','L_seq','R_seq','L_pos','R_pos','L_TM','R_TM','product_size']
        return pd.DataFrame(primers_list, columns=primers_col)

    def maketable(self):
        name_line = self.data_s.columns[9:]
        out_table_col = ['Chr','Pos','ID','Left','Right','L_TM','R_TM','Dif']
        for str_n in name_line:
            out_table_col.append('{}_Size'.format(str_n))

        out_list = []
        for i in range(len(self.data_s)):
            out_row = []
            out_row.append(self.data_s.at[self.data_s.index[i], '#CHROM'])
            out_row.append(self.data_s.at[self.data_s.index[i], 'POS'])
            out_row.append(self.primers.at[self.primers.index[i], 'name'])
            out_row.append(self.primers.at[self.primers.index[i], 'L_seq'])
            out_row.append(self.primers.at[self.primers.index[i], 'R_seq'])
            out_row.append(self.primers.at[self.primers.index[i], 'L_TM'])
            out_row.append(self.primers.at[self.primers.index[i], 'R_TM'])

            #Product size
            prod_size = int(self.primers.at[self.primers.index[i], 'product_size'])
            len_alt = len(self.data_s.at[self.data_s.index[i], 'ALT'])
            len_ref = len(self.data_s.at[self.data_s.index[i], 'REF'])
            out_row.append(abs(len_alt-len_ref)) #'Dif'

            for line in name_line:
                line_info = self.data_s.at[self.data_s.index[i], line]
                #ex. 1/1:0,6:6:18:270,18,0
                if line_info[0:3] == '0/0':
                    out_row.append(prod_size)
                elif line_info[0:3] == '1/1':
                    size = (prod_size + len_alt - len_ref)
                    out_row.append(size)
            out_list.append(out_row)
        self.out_table = pd.DataFrame(out_list, columns=out_table_col)

    def output(self):
        #1. output VCF
        out_vcf_name = '{}/{}.vcf'.format(self.dir, self.stem)
        with open(out_vcf_name, 'w') as o:
            for h in self.header:
                o.write('{}\n'.format(h))
            writer = csv.writer(o, delimiter='\t')
            writer.writerow(self.colnames)

        out_vcf_name_tmp = '{}.tmp'.format(out_vcf_name)
        self.data_s.to_csv(out_vcf_name_tmp, sep='\t', header=False, index=False)
        cmd1 = 'cat {} >> {}'.format(out_vcf_name_tmp, out_vcf_name)
        cmd2 = 'rm {}'.format(out_vcf_name_tmp)
        cmd1 = prepare_cmd(cmd1)
        cmd2 = prepare_cmd(cmd2)
        try:
            sbp.run(cmd1,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True, check=True)
            sbp.run(cmd2,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True, check=True)
        except sbp.CalledProcessError:
            print(time_stamp(),
              'Error occured writing vcf data.',
              flush=True)
            sys.exit(1) 

        #2. output primer data
        self.out_table.to_csv('{}/{}_primer_data.tsv'.format(self.dir, self.stem), sep='\t', header=True, index=False)

        print(time_stamp(), '{} markers were selected.'.format(len(self.data_s)), flush=True)

    def draw(self):
        out_vcf_name = '{}/{}.vcf'.format(self.dir, self.stem)
        out_png_name = '{}/{}.png'.format(self.dir, self.stem)

        vm = VisualizeMarker(out_vcf_name, out_png_name, self.fai)
        vm.run()
    

def main():
    print(time_stamp(), 'MKSelect started.', flush=True)

    prog = MKSelect(args)
    prog.command()
    prog.readvcf()
    prog.filtervcf()
    prog.maketable()
    prog.output()
    prog.draw()

    print(time_stamp(), 'MKSelect successfully finished.\n', flush=True)

if __name__ == '__main__':
    main()
    



import csv
import subprocess as sbp
import sys
from mkdesigner.utils import call_log, prepare_cmd, read_vcf, time_stamp

class RemoveTooLargeIndel(object):
    def __init__(self, args):
        self.args = args
        self.out = args.middir
        self.vcf = '{}/vcf_1st/Merged_filtered_variants.vcf'.format(self.out)
        self.outvcf = '{}/vcf_1st/Merged_filtered_variants_too_large_indel_removed.vcf'.format(self.out)
        
        self.maxsize = 100

    def run(self):
        print(time_stamp(),
              'Removing too large indels.',
              flush=True)
        
        #Read input VCF
        vcf_list = read_vcf(self.vcf)
        header = vcf_list[0]
        colnames = vcf_list[1]
        data = vcf_list[2]


        #VCF format: [0]#CHROM, [1]POS, [2]ID, [3]REF, [4]ALT, [5]QUAL, [6]FILTER, [7]INFO, [8]FORMAT	
        #'selected' is list of VCF rows filtered by following conditions.
        selected = []
        for i in range(0,len(data)):
            flag = 0
            ref = data[i][3].split(',')
            alt = data[i][4].split(',')
            for j in range(0,len(ref)):
                if len(ref[j]) > self.maxsize:
                    flag = 1
            for j in range(0,len(alt)):
                if len(alt[j]) > self.maxsize:
                    flag = 1
            if flag == 1:
                continue

            selected.append(data[i])

        with open(self.outvcf, 'w') as o:
            for h in header:
                o.write('{}\n'.format(h))
            writer = csv.writer(o, delimiter='\t')
            writer.writerow(colnames)
            writer.writerows(selected)
        
        cmd1 = 'gatk IndexFeatureFile -I {} \
                >> {}/log/gatk.log 2>&1'.format(self.outvcf, self.out)
        cmd1 = prepare_cmd(cmd1)
        try:
            sbp.run(cmd1,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True,
                    check=True)
        except sbp.CalledProcessError:
            call_log(self.out, 'gatk', cmd1)
            sys.exit(1)

        print(time_stamp(),
              '{} variants are selected from {} candidates.'.format(len(selected), len(data)),
              flush=True)
        print(time_stamp(),
              'Done.',
              flush=True)

        



#!/usr/bin/env python3

from mkdesigner.addprimertovcf import AddPrimerToVcf
from mkdesigner.params import Params
from mkdesigner.refindex import RefIndex
from mkdesigner.selectvariants import SelectVariants

pm = Params('mkprimer')
args = pm.set_options()

import os
import sys
import subprocess as sbp
from mkdesigner.utils import time_stamp, prepare_cmd

class MKPrimer(object):
    def __init__(self, args):
        pm.mkprimer_check_args(args)
        self.args = args
        self.ref = args.ref
        self.vcf = args.vcf
        self.name1 = args.name1
        self.name2 = args.name2
        self.type = args.type
        self.target = args.target
        self.cpu = args.cpu
        self.mindep = args.mindep
        self.maxdep = args.maxdep
        self.scope = args.search_span
        self.margin = args.margin
        self.primer_min_size = args.primer_min_size

        #240626 modified
        #devided output directory to {output}_mkprimer/ and {output}_mkprimer/intermediate/ 
        self.stem = args.output
        self.dir = '{}_mkprimer'.format(self.stem)
        self.middir = '{}_mkprimer/intermediate'.format(self.stem)

        #241023 added
        #set a limit number to consider primers
        self.limit = args.limit
        self.timeout = args.blast_timeout

    def mkdir(self):
        os.mkdir('{}'.format(self.dir))
        os.mkdir('{}'.format(self.middir))
        os.mkdir('{}/log'.format(self.middir))
        os.mkdir('{}/ref'.format(self.middir))
        os.mkdir('{}/vcf'.format(self.middir))
        os.mkdir('{}/intermediate'.format(self.middir))

    def command(self):
        #Output command info
        command = ' '.join(sys.argv)
        fn = '{}/command.txt'.format(self.dir)
        with open(fn, 'w') as f:
            f.write('{}\n'.format(command))

    def mvinputfiles(self):
        cmd1 = 'cp {} {}/ref/'.format(self.ref, self.middir)
        cmd2 = 'cp {} {}/vcf/'.format(self.vcf, self.middir)
        cmd1 = prepare_cmd(cmd1)
        cmd2 = prepare_cmd(cmd2)
        try:
            sbp.run(cmd1, stdout=sbp.DEVNULL, stderr=sbp.DEVNULL,
                    shell=True, check=True)
            sbp.run(cmd2, stdout=sbp.DEVNULL, stderr=sbp.DEVNULL,
                    shell=True, check=True)
        except sbp.CalledProcessError:
            print('Error occored at moving input files to the working directory.',flush=True)
            sys.exit(1)

    def refindex(self):
        ri = RefIndex(self)
        ri.run()

    def selectvariants(self):
        ss = SelectVariants(self)
        ss.run()

    def addprimertovcf(self):
        ad = AddPrimerToVcf(self)
        ad.run()

    def mvoutputfiles(self):
        output_vcf = '{}/vcf/{}_selected_primer_added.vcf'.format(self.middir, os.path.splitext(os.path.basename(self.vcf))[0])
        output_fai = '{}/ref/{}.fai'.format(self.middir, os.path.basename(self.ref))
        cmd1 = 'cp {} {}/{}_primer_added.vcf'.format(output_vcf, self.dir, self.stem)
        cmd2 = 'cp {} {}/for_draw.fai'.format(output_fai, self.dir)
        cmd1 = prepare_cmd(cmd1)
        cmd2 = prepare_cmd(cmd2)
        try:
            sbp.run(cmd1, stdout=sbp.DEVNULL, stderr=sbp.DEVNULL,
                    shell=True, check=True)
            sbp.run(cmd2, stdout=sbp.DEVNULL, stderr=sbp.DEVNULL,
                    shell=True, check=True)
        except sbp.CalledProcessError:
            print('Error occored at moving output files.',flush=True)
            sys.exit(1)

def main():
    print(time_stamp(), 'MKPrimer started.', flush=True)

    prog = MKPrimer(args)
    prog.mkdir()
    prog.command()
    prog.mvinputfiles()
    prog.refindex()
    prog.selectvariants()
    prog.addprimertovcf()
    prog.mvoutputfiles()
    
    print(time_stamp(), 'MKPrimer successfully finished.\n', flush=True)

if __name__ == '__main__':
    main()
    


__version__ = "0.5.3"
__copyright__    = 'Copyright (C) 2023 CHIGIRA Koki'
__license__      = 'MIT License'
__author__       = 'CHIGIRA Koki'
__author_email__ = 'kyoujin2009kutar@gmail.com'
__url__          = 'http://github.com/KChigira/mkdesigner'



#!/usr/bin/env python
from setuptools import setup, find_packages # type: ignore
from mkdesigner.__init__ import __version__

setup(name='mkdesigner',
      version=__version__,
      description='Genome-wide design of markers for PCR-based genotyping from NGS data.',
      author='Koki Chigira',
      author_email='kyoujin2009kutar@gmail.com',
      url='https://github.com/KChigira/mkdesigner/',
      license='MIT',
      packages=find_packages(),
      install_requires=[
        'pandas',
        'matplotlib',
        'timeout-decorator',
      ],
      entry_points={'console_scripts': [
            'mkvcf = mkdesigner.mkvcf:main',
            'mkprimer = mkdesigner.mkprimer:main',
            'mkselect = mkdesigner.mkselect:main',
            ]
      }
    )



import sys
import subprocess as sbp
from mkdesigner.utils import time_stamp, prepare_cmd, call_log


class MergeVcf(object):

    def __init__(self, args, is2nd=False):
        self.args = args
        self.out = args.middir
        self.name = args.name

        if is2nd:
            self.vcfdir = '{}/vcf_2nd/'.format(self.out)
        else: #1st
            self.vcfdir = '{}/vcf_1st/'.format(self.out)

    def run(self):
        print(time_stamp(),
              'Merging vcf files.',
              flush=True)
        
        vcf_all = [self.vcfdir + s + '_filtered_variants.vcf.gz' for s in self.name]
        vcf_con = ' '.join(vcf_all)

        with open('{}sample_names.txt'.format(self.vcfdir), 'w') as o:
            for j in range(len(self.name)):
                o.write('{}\n'.format(self.name[j]))

        cmd1 = 'bcftools merge -o {}Merged_filtered_variants_raw.vcf.gz -O z --force-samples \
                {} \
                >> {}/log/bcftools.log \
                2>&1'.format(self.vcfdir, vcf_con, self.out)
        cmd1_1 = 'bcftools reheader -o {}Merged_filtered_variants.vcf.gz -s {}sample_names.txt \
                {}Merged_filtered_variants_raw.vcf.gz \
                >> {}/log/bcftools.log \
                2>&1'.format(self.vcfdir, self.vcfdir, self.vcfdir, self.out)

        #cmd2 = 'bcftools index {}Merged_filtered_variants.vcf.gz \
        #        >> {}/log/bcftools.log 2>&1'.format(self.vcfdir, self.out)
        cmd2 = 'gatk IndexFeatureFile -I {}Merged_filtered_variants.vcf.gz \
                >> {}/log/gatk.log 2>&1'.format(self.vcfdir, self.out)

        cmd3 = 'gunzip -c {}Merged_filtered_variants.vcf.gz > {}Merged_filtered_variants.vcf'.format(self.vcfdir, self.vcfdir)

        cmd1 = prepare_cmd(cmd1)
        cmd1_1 = prepare_cmd(cmd1_1)
        cmd2 = prepare_cmd(cmd2)
        cmd3 = prepare_cmd(cmd3)

        try:
            sbp.run(cmd1,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True,
                    check=True)
        except sbp.CalledProcessError:
            call_log(self.out, 'bcftools', cmd1)
            sys.exit(1)

        try:
            sbp.run(cmd1_1,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True,
                    check=True)
        except sbp.CalledProcessError:
            call_log(self.out, 'bcftools', cmd1_1)
            sys.exit(1)

        try:
            sbp.run(cmd2,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True,
                    check=True)
        except sbp.CalledProcessError:
            call_log(self.out, 'gatk', cmd2)
            sys.exit(1)

        try:
            sbp.run(cmd3,
                    stdout=sbp.DEVNULL,
                    stderr=sbp.DEVNULL,
                    shell=True,
                    check=True)
        except sbp.CalledProcessError:
            call_log(self.out, 'gunzip', cmd3)
            sys.exit(1)

        print(time_stamp(),
              'Done.',
              flush=True)




import csv
import sys
import os
from mkdesigner.utils import read_vcf, time_stamp

class SelectVariants(object):
    def __init__(self, args):
        self.args = args
        self.out = args.middir
        self.name1 = args.name1
        self.name2 = args.name2
        self.mindep = args.mindep
        self.maxdep = args.maxdep
        self.target = args.target

        self.refidx = '{}/ref/{}.fai'.format(self.out, os.path.basename(args.ref))
        self.vcf = '{}/vcf/{}'.format(self.out, os.path.basename(args.vcf))
        self.outvcf = '{}/vcf/{}_selected.vcf'.format(self.out, os.path.splitext(os.path.basename(args.vcf))[0])

    def run(self):
        print(time_stamp(),
              'Selecting variants which meet the conditions.',
              flush=True)
        
        #Read input VCF
        vcf_list = read_vcf(self.vcf)
        header = vcf_list[0]
        colnames = vcf_list[1]
        data = vcf_list[2]
        header.append('##selectsnp.py_{}'.format(time_stamp()))

        #Get column numbers of focused varieties.
        id_1 = [] #List of Integer
        id_2 = [] #List of Integer
        try:
            for n in self.name1:
                id_1.append(colnames.index(n))
            for n in self.name2:
                id_2.append(colnames.index(n))
        except ValueError:
            print(time_stamp(), '!!ERROR!! Specified line does not exist in the VCF file\n', flush=True)
            sys.exit(1)

        #Read input reference fasta.fai
        chrnames = []
        chrlens = []
        with open(self.refidx, 'r') as r:
            reader = csv.reader(r, delimiter='\t')
            for row in reader:
                chrnames.append(row[0])
                chrlens.append(row[1])
        
        #240626 added. 
        #Select variants if target position is designated.
        if self.target is not None:
            self.target_chr = []
            self.target_start = []
            self.target_end = []
            for h in range(len(self.target)):
                tmp = self.target[h].split(':')
                if len(tmp) != 2:
                    print(time_stamp(), 'input of "--target" is invalid.', flush=True)
                    sys.exit(1)
                tmp2 = tmp[1].split('-')
                if len(tmp2) != 2:
                    print(time_stamp(), 'input of "--target" is invalid.', flush=True)
                    sys.exit(1)
                self.target_chr.append(tmp[0])
                self.target_start.append(int(tmp2[0]))
                self.target_end.append(int(tmp2[1]))

        #VCF format: [0]#CHROM, [1]POS, [2]ID, [3]REF, [4]ALT, [5]QUAL, [6]FILTER, [7]INFO, [8]FORMAT	
        #'selected' is list of VCF rows filtered by following conditions.
        selected = []
        for i in range(0,len(data)):
            if self.target is not None:
                flag = 0
                for j in range(len(self.target_chr)):
                    if data[i][0] == self.target_chr[j] and \
                    int(data[i][1]) > self.target_start[j] and \
                    int(data[i][1]) < self.target_end[j]:
                        flag = 1
                if flag == 0: #remove variants which are not target
                    continue
                
            if len(data[i][4].split(',')) > 1:
                continue # remove multi allelic variants
            if self.args.type == 'SNP':
                if abs(len(data[i][3]) - len(data[i][4])) != 0:
                    continue #For SNP marker, remove INDEL.
            elif self.args.type == 'INDEL':
                if abs(len(data[i][3]) - len(data[i][4])) == 0:
                    continue #For INDEL marker, remove SNP.
                if abs(len(data[i][3]) - len(data[i][4])) > 50:
                    continue #INDEL larger than 50bp are removed because they are suspecible.
            if data[i][6] != 'PASS':
                continue # remove low quality variants
            if int(data[i][1]) <= self.args.scope:
                continue # remove variants too close to 5' terminal
            if int(data[i][1]) + self.args.scope > int(chrlens[chrnames.index(data[i][0])]):
                continue # remove variants too close to 3' terminal

            #Check haplotype calling format.
            format = data[i][8].split(':')
            geno_1 = [] #List of haplotype calling data of varities 1
            geno_2 = [] #List of haplotype calling data of varities 2
            for j in id_1:
                geno_1.append(data[i][j].split(':'))
            for j in id_2:
                geno_2.append(data[i][j].split(':'))
            try:
                gt_index = format.index('GT')
                dp_index = format.index('DP')
                gt_1 = [] #List of 'GT' value for varieties 1
                          #(genotype, ex. "0/0""0/1""1/1")
                dp_1 = [] #List of 'DP' value for varieties 1
                          #(depth, integer)
                gt_2 = [] #List of 'GT' value for varieties 2
                dp_2 = [] #List of 'DP' value for varieties 2
                for g in geno_1:
                    gt_1.append(g[gt_index])
                    dp_1.append(int(g[dp_index]))
                for g in geno_2:
                    gt_2.append(g[gt_index])
                    dp_2.append(int(g[dp_index]))
            except (ValueError, IndexError):
                continue #Maybe lack of GT or DP data.
            
            #Check combination of genotypes
            if len(set(gt_1)) != 1 or len(set(gt_2)) != 1 \
                                   or gt_1[0] == gt_2[0]:
                #if genotypes of all varieties_1 are same,
                #length of set(gt_1) == 1 because the set data structure
                #does not allow duplicates.
                continue

            #Check homozygous or not
            gt_1_allele = str(gt_1[0]).replace('|', '/').split('/')
            gt_2_allele = str(gt_2[0]).replace('|', '/').split('/')
            try:
                if int(gt_1_allele[0]) != int(gt_1_allele[1]):
                    continue
                if int(gt_2_allele[0]) != int(gt_2_allele[1]):
                    continue
            except ValueError:
                continue #Sometimees haplotype like "./." exist.

            #Check depth
            if not (min(dp_1) >= self.mindep and max(dp_1) <= self.maxdep 
                    and min(dp_2) >= self.mindep and max(dp_2) <= self.maxdep):
                continue

            #Get span where we can design primers 
            chrend = int(chrlens[chrnames.index(data[i][0])]) #230921 modified
            if i == 0:
                chr_3 = ['', data[i][0], data[i+1][0]]
            elif i == len(data) - 1:
                chr_3 = [data[i-1][0], data[i][0], '']
            else:
                chr_3 = [data[i-1][0], data[i][0], data[i+1][0]]

            if chr_3[0] != chr_3[1] and chr_3[1] == chr_3[2]:
                pos_3 = [0, int(data[i][1]), int(data[i+1][1])]
            elif chr_3[0] == chr_3[1] and chr_3[1] != chr_3[2]:
                pos_3 = [int(data[i-1][1]), int(data[i][1]), chrend]
            elif chr_3[0] != chr_3[1] and chr_3[1] != chr_3[2]:
                pos_3 = [0, int(data[i][1]), chrend]
            else:
                pos_3 = [int(data[i-1][1]), int(data[i][1]), int(data[i+1][1])]

            data[i][7] = '{};SPAN={},{}'.format(data[i][7], pos_3[1]-pos_3[0], pos_3[2]-pos_3[1])

            selected.append(data[i])

        with open(self.outvcf, 'w') as o:
            for h in header:
                o.write('{}\n'.format(h))
            writer = csv.writer(o, delimiter='\t')
            writer.writerow(colnames)
            writer.writerows(selected)
        
        print(time_stamp(),
              '{} variants are selected from {} candidates.'.format(len(selected), len(data)),
              flush=True)
        print(time_stamp(),
              'Done.',
              flush=True)

        